(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_io.ml 239 2010-06-23 16:49:03Z gerd $ *)

open Plasma_rpcapi_aux

class type record_reader =
object
  method pos_in : int
  method input_record : unit -> string
  method peek_record : unit -> string
  method close_in : unit -> unit
  method abort : unit -> unit
end


class type record_writer =
object
  method pos_out : int
  method bpos_out : int64
  method output_record : string -> unit
  method flush : unit -> unit
  method close_out : unit -> unit
  method abort : unit -> unit
end


let normal_mem_limit = 64 * 1024 * 1024  (* buffer up to 64 M  *)
let tight_mem_limit  = 16 * 1024 * 1024  (* buffer up to 16 M  *)

let zero_mem =
  Bigarray.Array1.create Bigarray.char Bigarray.c_layout 0

let read_file c name index len : record_reader =
  let () =
    if index < 0L || len <= 0L then invalid_arg "Mapred_io.read_file" in
  let bsize = Plasma_client.blocksize c in
  let bsizeL = Int64.of_int bsize in
  let shm_mng = Plasma_client.shm_manager c in
  let shm_cond = shm_mng # shm_condition in
  let shm_id = shm_mng # new_id() in
  let mem_limit = 
    match shm_cond with
      | `Normal -> max bsize normal_mem_limit
      | `Tight -> max bsize tight_mem_limit in
  let (inode,ii) =
    Plasma_client.retry
      name
      10   (* TODO: configurable *)
      (fun () ->
	 let trans = Plasma_client.start c in
	 try
	   let inode = Plasma_client.lookup trans name in
	   let ii = Plasma_client.get_inodeinfo trans inode in
	   Plasma_client.abort trans;
	   (inode,ii)
	 with
	   | error ->
	       Plasma_client.abort trans;
	       raise error
      )
      () in
  let () = Plasma_client.abort_cluster c in
  let max_index = 
    if ii.eof = 0L then (-1L) else Int64.div (Int64.pred ii.eof) bsizeL in
  let () =
    if index > Int64.sub (Int64.succ max_index) len then
      failwith "Mapred_io.read_file: cannot read beyond eof" in
  let end_index =            (* logical end *)
    Int64.add index len in
  let hard_end_index =       (* logical end + next block if possible *)
    min (Int64.succ end_index) (Int64.succ max_index) in
  let hard_len =
    Int64.sub hard_end_index index in
  let buf_limit = max 1 (mem_limit / bsize) in
  let buf_numL = min hard_len (Int64.of_int buf_limit) in
  let buf_num = Int64.to_int buf_numL in
  let buf_dim = buf_num * bsize in
  let (buf_fd,buf_name) = Plasma_util.get_shm_fd "mapred" 0 in
  let bufs =
    try
      Netsys_posix.shm_unlink buf_name;
      Netlog.logf `Debug "read_file (pid %d): truncate fd=%Ld eof=%d"
	(Unix.getpid()) (Netsys.int64_of_file_descr buf_fd) buf_dim;
      Plasma_util.enlarge_shm buf_fd buf_dim;
      shm_mng # book_allocation shm_id buf_dim;
      Plasma_util.map_shm buf_fd 0L buf_dim
    with
      | error -> Unix.close buf_fd; raise error in
  let buf_end = ref 0 in        (* end position in bufs in bytes *)
  let buf_count = ref 0 in      (* how many blocks we have *)
  let buf_eof = ref None in     (* the (logical) eof position if reached *)
  let buf_index = ref index in  (* the block the first buffer holds *)
  let buf_cursor = ref 0 in     (* as byte position in bufs *)
  let buf_eof_reached = ref false in

  let find_lf (bufs : Netsys_mem.memory) pos end_pos =
    assert(pos <= buf_dim && end_pos <= buf_dim);
    Platform.find_lf bufs pos end_pos
(*
    let pos = ref pos in
    while !pos < end_pos && Bigarray.Array1.unsafe_get bufs !pos <> '\n' do
      incr pos
    done;
    if !pos = end_pos then raise Not_found;
    !pos
 *)
  in

object(self)

  val mutable pos_in = 0
  val mutable cur_rec = None
  val mutable closed = false

  initializer
    self # fill_buffers index;
    assert(!buf_count > 0);
    if index > 0L then (
      try
	let p = find_lf bufs 0 !buf_end in
	if p >= bsize then raise Not_found;
	buf_cursor := p+1
      with
	| Not_found ->
	    failwith "Mapred_io.read_file: record too large"
    )

  method pos_in = pos_in

  method peek_record() =
    if closed then failwith "Mapred_io.read_file: closed";
    match cur_rec with
      | None -> 
	  let line = self # input_record_rest "" in
	  cur_rec <- Some line;
	  line
      | Some r -> r

  method input_record() =
    if closed then failwith "Mapred_io.read_file: closed";
    match cur_rec with
      | None -> 
(*
    if pos_in >= 603950 then (
      Printf.printf
	"pos_in=%d buf_cursor=%d buf_end=%d buf_eof=%s buf_eof_reached=%b\n%!"
	pos_in !buf_cursor !buf_end
	( match !buf_eof with None -> "none" | Some e -> string_of_int e)
	!buf_eof_reached
    );
 *)
	  let line = self # input_record_rest "" in
	  cur_rec <- None;
	  pos_in <- pos_in + 1;
	  line
      | Some r -> 
	  cur_rec <- None;
	  pos_in <- pos_in + 1;
	  r

  method private input_record_rest line0 =
    (* prefix result with line0 *)
    if !buf_eof_reached then raise End_of_file;
    assert(!buf_cursor <= !buf_end);
    let l0 = String.length line0 in
    try
      let p = find_lf bufs !buf_cursor !buf_end in
      let l = l0 + p - !buf_cursor in
      if l >= bsize then
	failwith "Mapred_io.read_file: record too large";
      let line = String.create l in
      if l0 > 0 then
	String.blit line0 0 line 0 l0;
      Netsys_mem.blit_memory_to_string bufs !buf_cursor line l0 (l-l0);
      buf_cursor := p+1;
      (* maybe we have exceeded the soft limit already *)
      ( match !buf_eof with
	  | None -> ()
	  | Some eof ->
	      if p >= eof then
		buf_eof_reached := true
      );
      line
    with
      | Not_found ->
	  if l0 > 0 || !buf_end - !buf_cursor >= bsize then
	    failwith "Mapred_io.read_file: record too large";
	  (* we have to look into the next block if it exists: *)
	  let l0 = !buf_end - !buf_cursor in
	  let line0 = String.create l0 in
	  Netsys_mem.blit_memory_to_string bufs !buf_cursor line0 0 l0;
	  match !buf_eof with
	    | Some _ ->
		buf_eof_reached := true;
		(* maybe the last line did not end with LF *)
		if l0 = 0 then raise End_of_file;
		line0
	    | None ->
		(* load the following blocks: *)
		self # fill_buffers 
		  (Int64.add !buf_index (Int64.of_int !buf_count));
		if !buf_count = 0 then (
		  (* strange case, but better handle that *)
		  buf_eof_reached := true;
		  (* maybe the last line did not end with LF *)
		  if l0 = 0 then raise End_of_file;
		  line0
		) else
		  self # input_record_rest line0
		    

  method private fill_buffers i =
    (* Fill buffers so that the first one has index i *)
    
    let max_len_blocks = 
      Int64.to_int (min (Int64.sub hard_end_index i) buf_numL) in
    let len = max_len_blocks * bsize in  (* rounded up to blocks *)
    let lenL = Int64.of_int len in
    let pos = Int64.mul i bsizeL in

    Netlog.logf `Debug
      "Mapred_io.read_file: copy_out inode=%Ld pos=%Ld len=%d"
      inode pos len;

    Plasma_client.retry
      name
      10 (* TODO *)
      (fun () ->
	 ignore(Unix.lseek buf_fd 0 (Unix.SEEK_SET));
	 Plasma_client.copy_out c inode pos buf_fd lenL
      )
      ();

    let n = Int64.to_int (min lenL (Int64.sub ii.eof pos)) in

    (* copy_out does not look at eof, but dumbly tries to get the
       blocks
     *)
    let eof = Int64.add pos lenL >= ii.eof in

    buf_index := i;
    buf_end := n;
    buf_count := if len=0 then 0 else (len - 1) / bsize + 1;
    buf_eof := 
      if eof || Int64.add i (Int64.of_int !buf_count) >= end_index then
	let m = Int64.to_int (Int64.sub end_index i) * bsize in
	if m = n && not eof then
	  None  (* otherwise we wouldn't read the next block *)
	else
	  Some n
      else
	None;
    buf_cursor := 0;

    Netlog.logf `Debug
      "Mapred_io.read_file: new buffer: index=%Ld end=%d count=%d eof=%s"
      !buf_index !buf_end !buf_count
      (match !buf_eof with None -> "none" | Some n -> string_of_int n);

    Plasma_client.abort_cluster c

  method close_in() =
    if not closed then (
      Plasma_client.abort_cluster c;
      Unix.ftruncate buf_fd 0;
      Unix.close buf_fd;
      Netsys_mem.memory_unmap_file bufs;
      shm_mng # book_allocation shm_id 0;
      closed <- true
    )

  method abort() =
    self # close_in()

end


let read_multiple readers =
  let readers = ref readers in
object(self)
  val mutable pos_in = 0
  val mutable cur_rec = None
  val mutable cur_rd = None

  method pos_in = pos_in

  method peek_record() =
    match cur_rec with
      | None -> 
	  let line = self#next_record() in
	  cur_rec <- Some line;
	  line
      | Some r -> r

  method input_record() =
    match cur_rec with
      | None ->
	  let line = self#next_record() in
	  pos_in <- pos_in + 1;
	  line
      | Some r ->
	  cur_rec <- None;
	  pos_in <- pos_in + 1;
	  r

  method private next_record() =
    match cur_rd with
      | Some rd ->
	  ( try
	      rd # input_record()
	    with
	      | End_of_file ->
		  rd # close_in();
		  cur_rd <- None;
		  self # next_reader()
	  )
      | None ->
	  self # next_reader()

  method private next_reader() =
    match !readers with
      | first :: rest ->
	  readers := rest;
	  cur_rd <- Some (first());
	  self # next_record()
      | [] ->
	  raise End_of_file

  method close_in() =
    match cur_rd with
      | Some rd ->
	  rd # close_in()
      | None ->
	  ()

  method abort() =
    match cur_rd with
      | Some rd ->
	  rd # abort()
      | None ->
	  ()

end


let write_file c name =
  let bsize = Plasma_client.blocksize c in
  let shm_mng = Plasma_client.shm_manager c in
  let shm_cond = shm_mng # shm_condition in
  let shm_id = shm_mng # new_id() in
  let mem_limit = 
    match shm_cond with
      | `Normal -> max bsize normal_mem_limit
      | `Tight -> max bsize tight_mem_limit in
  (* let bsizeL = Int64.of_int bsize in *)
  let inode =
    Plasma_client.retry
      name
      10 (* TODO *)
      (fun () ->
	 let trans = Plasma_client.start c in
	 try
	   let inode = Plasma_client.lookup trans name in
	   let ii = Plasma_client.get_inodeinfo trans inode in
	   let ii' = { ii with eof = 0L } in
	   Plasma_client.set_inodeinfo trans inode ii';
	   Plasma_client.commit trans;
	   inode
	 with
	   | error ->
	       Plasma_client.abort trans; raise error
      )
      () in
  let () = Plasma_client.abort_cluster c in
  let buf_num = max 1 (mem_limit / bsize) in
  (* let buf_numL = Int64.of_int buf_num in *)
  let buf_len = buf_num * bsize in
  let (buf_fd,buf_name) = Plasma_util.get_shm_fd "mapred" 0 in
  let bufs =
    try
      Netsys_posix.shm_unlink buf_name;
      Netlog.logf `Debug "write_file (pid %d): truncate fd=%Ld eof=%d"
	(Unix.getpid()) (Netsys.int64_of_file_descr buf_fd) buf_len;
      Plasma_util.enlarge_shm buf_fd buf_len;
      shm_mng # book_allocation shm_id buf_len;
      Plasma_util.map_shm buf_fd 0L buf_len
    with
      | error -> Unix.close buf_fd; raise error in
  let buf_end = ref 0 in        (* end position in bufs in bytes *)
  let buf_pos = ref 0L in       (* byte position in file of first buf byte *)
object(self)
  val mutable pos_out = 0
  val mutable bpos_out = 0L
  val mutable closed = false

  method pos_out = pos_out
  method bpos_out = bpos_out

  method output_record line =
    (* we do not protect against having LF in line. User's responsibility *)
    if closed then failwith "Mapred_io.write_file: closed";
    let l = String.length line in
    if l >= bsize then
      failwith "Mapred_io.write_file: record too large";
    let p = ref 0 in
    while !p < l+1 do
      let space = buf_len - !buf_end in
      let n =
	if !p < l then (
	  let k = min (l - !p) space in
	  Netsys_mem.blit_string_to_memory line !p bufs !buf_end k;
	  k
	) else (
	  bufs.{ !buf_end } <- '\n';
	  1
	) in
      buf_end := !buf_end + n;
      p := !p + n;
      if !buf_end = buf_len then
	self # flush()
    done;
    bpos_out <- Int64.add bpos_out (Int64.of_int (l + 1))

  method flush() =
    if closed then failwith "Mapred_io.write_file: closed";
    Netlog.logf `Debug "Mapred_io.flush: inode=%Ld pos=%Ld len=%d"
      inode !buf_pos !buf_end;

    Plasma_client.retry
      name
      10 (* TODO *)
      (fun () ->
	 ignore(Unix.lseek buf_fd 0 (Unix.SEEK_SET));
	 Plasma_client.copy_in
	   c inode !buf_pos buf_fd (Int64.of_int !buf_end) `Star
      )
      ();
    (* copy_in works only with multiples of bsize. So if we did not match
       this, retain the end of the buffer
     *)
    let tail = !buf_end mod bsize in
    if tail > 0 then (
      let n = !buf_end - tail in
      Bigarray.Array1.blit
	(Bigarray.Array1.sub bufs (!buf_end - tail) tail)
	(Bigarray.Array1.sub bufs 0 tail);
      buf_pos := Int64.add !buf_pos (Int64.of_int n);
      buf_end := tail
    )
    else (
      buf_pos := Int64.add !buf_pos (Int64.of_int !buf_end);
      buf_end := 0;
    );

    Plasma_client.abort_cluster c 
    
  method close_out() =
    if not closed then (
      self # flush();
      Unix.ftruncate buf_fd 0;
      Unix.close buf_fd;
      Netsys_mem.memory_unmap_file bufs;
      shm_mng # book_allocation shm_id 0;
      closed <- true
    )

  method abort() =
    if not closed then (
      (* no flush here *)
      Unix.close buf_fd;
      Netsys_mem.memory_unmap_file bufs;
      shm_mng # book_allocation shm_id 0;
      closed <- true
    )
end


let write_multiple c prefix limit create =
  let n = ref 0 in
  let current = ref None in
  let switch() = 
    let name = create prefix !n in
    incr n;
    let w = write_file c name in
    current := Some w;
    w in
object(self)
  val mutable pos_out = 0
  val mutable bpos_out = 0L

  method pos_out = pos_out
  method bpos_out = bpos_out

  method output_record line =
    let w =
      match !current with
	| None ->
	    switch()
	| Some w ->
	    let p =
	      Int64.add w#bpos_out (Int64.of_int (String.length line + 1))  in
	    if p >= limit then (
	      w # close_out();
	      switch()
	    )
	    else
	      w in
    w # output_record line

  method flush() =
    match !current with
      | None -> ()
      | Some w -> w # flush()

  method close_out () =
    match !current with
      | None -> ()
      | Some w -> w # close_out()

  method abort() =
    match !current with
      | None -> ()
      | Some w -> w # abort()
end


let create_file ?(repl=0) c name =
  Plasma_client.retry
    name
    10 (* TODO *)
    (fun () ->
       let trans = Plasma_client.start c in
       ( try
	   let ii = Plasma_client.regular_ii c 0o666 in
	   let ii = { ii with replication = repl } in
	   let _ = Plasma_client.create_file trans name ii in
	   Plasma_client.commit trans;
	   Netlog.logf `Debug "create_file name=%s" name
	 with
	   | error ->
	       Plasma_client.abort trans;
	       raise error
       )
    )
    ()


let delete_file c name =
  Plasma_client.retry
    name
    10 (* TODO *)
    (fun () ->
       let trans = Plasma_client.start c in
       ( try
	   let _ = Plasma_client.unlink trans name in
	   Plasma_client.commit trans;
	   Netlog.logf `Debug "delete_file name=%s" name
	 with
	   | error ->
	       Plasma_client.abort trans;
	       raise error
       )
    )
    ()


let file_blocks c name =
  Plasma_client.retry
    name
    10 (* TODO *)
    (fun () ->
       let trans = Plasma_client.start c in
       ( try
	   let bsize = Plasma_client.blocksize c in
	   let bsizeL = Int64.of_int bsize in
	   let ii = 
	     Plasma_client.get_inodeinfo trans (Plasma_client.lookup trans name) in
	   let blocks =
	     if ii.eof = 0L then 0L 
	     else
	       Int64.succ (Int64.div (Int64.pred ii.eof) bsizeL) in
	   Plasma_client.commit trans;
	   blocks
	 with
	   | error ->
	       Plasma_client.abort trans;
	       raise error
       )
    )
    ()


