(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_commit.mli 239 2010-06-23 16:49:03Z gerd $ *)

(** Two-phase commit protocol *)


val commit_e : (bool -> unit) ->
               Nn_db.modification Queue.t ->
               Nn_state.shared_state_t ->
                 bool Uq_engines.engine
  (** [commit_e post_commit journal shared_state]:

      Writes the modifications to the local database and all slaves,
      and commits the update. The bool arg says whether the commit
      was successful. If [false], the transaction is guaranteed to be
      rolled back. The commit engine may also enter an error state.
      This is considered as a critical violation of the protocol.

      If the queue of modification is empty, the local transaction is
      committed, and no remote operations are involved. (This means
      that the user only submitted read requests and no writes.)

      If there are modifications, a new revision string is created
      and appended to the journal before anything is written.

      Commits are serialized (using [finish_transaction_serialized]).

      As side effects, the function may disable slaves, and even
      (self-)disable the coordinator.

      The function also invokes [post_commit] when the commit or
      rollback is done. The bool arg is the same as in the result.
   *)


val trans_e :  Nn_state.shared_state_t ->
               Nn_state.view_t list -> 
              'a Uq_engines.engine -> 
                'a Uq_engines.final_state Uq_engines.engine
  (** [let e' = trans_e sh vl e]: Runs the engine [e] until it is finished,
      and if [e] was successful, the modifications of the views [vl]
      are taken as the modifications to be done in the transaction,
      and the transaction is committed.

      If [e] is not successful,
      the views and the db transaction are rolled back. 

      The returned engine [e'] is [`Done(st)] when
      the transaction is over (after commit or rollback). [st] reflects the
      final state of [e].

      [e'] reaches an error state only if the commit fails (critical error).
   *)

